@@ -156,6 +156,13 @@ static int set_sys_acl_conn(const char *fname,
 
 	ret = SMB_VFS_SYS_ACL_SET_FD(smb_fname->fsp, acltype, theacl);
 
+	status = fd_close(smb_fname->fsp);
+	if (!NT_STATUS_IS_OK(status)) {
+		TALLOC_FREE(frame);
+		errno = map_errno_from_nt_status(status);
+		return -1;
+	}
+
 	TALLOC_FREE(frame);
 	return ret;
 }
@@ -169,6 +176,7 @@ static NTSTATUS init_files_struct(TALLOC_CTX *mem_ctx,
 {
 	struct smb_filename *smb_fname = NULL;
 	int fd;
+	int ret;
 	mode_t saved_umask;
 	struct files_struct *fsp;
 	struct files_struct *fspcwd = NULL;
@@ -222,13 +230,13 @@ static NTSTATUS init_files_struct(TALLOC_CTX *mem_ctx,
 	}
 	fsp_set_fd(fsp, fd);
 
-	status = vfs_stat_fsp(fsp);
-	if (!NT_STATUS_IS_OK(status)) {
+	ret = SMB_VFS_FSTAT(fsp, &smb_fname->st);
+	if (ret == -1) {
 		/* If we have an fd, this stat should succeed. */
 		DEBUG(0,("Error doing fstat on open file %s (%s)\n",
 			 smb_fname_str_dbg(smb_fname),
-			 nt_errstr(status) ));
-		return status;
+			 strerror(errno) ));
+		return map_nt_error_from_unix(errno);
 	}
 
 	fsp->file_id = vfs_file_id_from_sbuf(conn, &smb_fname->st);
@@ -274,7 +282,7 @@ static NTSTATUS set_nt_acl_conn(const char *fname,
 		DBG_ERR("init_files_struct failed: %s\n",
 			nt_errstr(status));
 		if (fsp != NULL) {
-			SMB_VFS_CLOSE(fsp);
+			fd_close(fsp);
 		}
 		TALLOC_FREE(frame);
 		return status;
@@ -285,7 +293,7 @@ static NTSTATUS set_nt_acl_conn(const char *fname,
 		DEBUG(0,("set_nt_acl_no_snum: fset_nt_acl returned %s.\n", nt_errstr(status)));
 	}
 
-	SMB_VFS_CLOSE(fsp);
+	fd_close(fsp);
 
 	TALLOC_FREE(frame);
 	return status;
@@ -332,6 +340,12 @@ static NTSTATUS get_nt_acl_conn(TALLOC_CTX *mem_ctx,
 			nt_errstr(status));
 	}
 
+	status = fd_close(smb_fname->fsp);
+	if (!NT_STATUS_IS_OK(status)) {
+		TALLOC_FREE(frame);
+		return status;
+	}
+
 	TALLOC_FREE(frame);
 
 	return status;
@@ -617,7 +631,7 @@ static PyObject *py_smbd_chown(PyObject *self, PyObject *args, PyObject *kwargs)
 		DBG_ERR("init_files_struct failed: %s\n",
 			nt_errstr(status));
 		if (fsp != NULL) {
-			SMB_VFS_CLOSE(fsp);
+			fd_close(fsp);
 		}
 		TALLOC_FREE(frame);
 		/*
@@ -630,13 +644,13 @@ static PyObject *py_smbd_chown(PyObject *self, PyObject *args, PyObject *kwargs)
 	ret = SMB_VFS_FCHOWN(fsp, uid, gid);
 	if (ret != 0) {
 		int saved_errno = errno;
-		SMB_VFS_CLOSE(fsp);
+		fd_close(fsp);
 		TALLOC_FREE(frame);
 		errno = saved_errno;
 		return PyErr_SetFromErrno(PyExc_OSError);
 	}
 
-	SMB_VFS_CLOSE(fsp);
+	fd_close(fsp);
 	TALLOC_FREE(frame);
 
 	Py_RETURN_NONE;
@@ -1041,6 +1055,13 @@ static PyObject *py_smbd_get_sys_acl(PyObject *self, PyObject *args, PyObject *k
 		return PyErr_SetFromErrno(PyExc_OSError);
 	}
 
+	status = fd_close(smb_fname->fsp);
+	if (!NT_STATUS_IS_OK(status)) {
+		TALLOC_FREE(frame);
+		PyErr_SetNTSTATUS(status);
+		return NULL;
+	}
+
 	py_acl = py_return_ndr_struct("samba.dcerpc.smb_acl", "t", acl, acl);
 
 	TALLOC_FREE(frame);
@@ -1209,7 +1230,7 @@ static PyObject *py_smbd_create_file(PyObject *self, PyObject *args, PyObject *k
 		DBG_ERR("init_files_struct failed: %s\n",
 			nt_errstr(status));
 	} else if (fsp != NULL) {
-		SMB_VFS_CLOSE(fsp);
+		fd_close(fsp);
 	}
 
 	TALLOC_FREE(frame);
