@@ -63,6 +63,7 @@ static PyObject *py_dnsp_DnssrvRpcRecord_get_list(struct dnsp_DnssrvRpcRecord *r
 	return py_dns_list;
 }
 
+
 static int py_dnsp_DnssrvRpcRecord_get_array(PyObject *value,
 					     TALLOC_CTX *mem_ctx,
 					     struct dnsp_DnssrvRpcRecord **records,
@@ -310,7 +311,7 @@ static PyObject *py_dsdb_dns_replace_by_dn(PyObject *self, PyObject *args)
 	werr = dns_common_replace(samdb,
 				  frame,
 				  dn,
-				  false, /* Not adding a record */
+				  false, /* Not adding a node */
 				  serial,
 				  records,
 				  num_records);
@@ -325,6 +326,81 @@ static PyObject *py_dsdb_dns_replace_by_dn(PyObject *self, PyObject *args)
 	Py_RETURN_NONE;
 }
 
+
+static PyObject *py_dsdb_dns_records_match(PyObject *self, PyObject *args)
+{
+	PyObject *py_recs[2];
+	struct dnsp_DnssrvRpcRecord *rec1;
+	struct dnsp_DnssrvRpcRecord *rec2;
+	size_t i;
+	bool type_correct;
+	bool match;
+
+	if (!PyArg_ParseTuple(args, "OO", &py_recs[0], &py_recs[1])) {
+		return NULL;
+	}
+
+	for (i = 0; i < 2; i++) {
+		type_correct = py_check_dcerpc_type(py_recs[i],
+						    "samba.dcerpc.dnsp",
+						    "DnssrvRpcRecord");
+		if (! type_correct) {
+			PyErr_SetString(PyExc_ValueError,
+					"DnssrvRpcRecord expected");
+			return NULL;
+		}
+	}
+
+	rec1 = (struct dnsp_DnssrvRpcRecord *)pytalloc_get_ptr(py_recs[0]);
+	rec2 = (struct dnsp_DnssrvRpcRecord *)pytalloc_get_ptr(py_recs[1]);
+
+	match = dns_record_match(rec1, rec2);
+	return PyBool_FromLong(match);
+}
+
+
+static PyObject *py_dsdb_dns_unix_to_dns_timestamp(PyObject *self, PyObject *args)
+{
+	uint32_t timestamp;
+	time_t t;
+	long long lt;
+
+	if (!PyArg_ParseTuple(args, "L", &lt)) {
+		return NULL;
+	}
+
+	t = lt;
+	if (t != lt) {
+		/* time_t is presumably 32 bit here */
+		PyErr_SetString(PyExc_ValueError, "Time out of range");
+		return NULL;
+	}
+	timestamp = unix_to_dns_timestamp(t);
+	return Py_BuildValue("k", (unsigned long) timestamp);
+}
+
+static PyObject *py_dsdb_dns_timestamp_to_nt_time(PyObject *self, PyObject *args)
+{
+	unsigned long long timestamp;
+	NTSTATUS status;
+	NTTIME nt;
+	if (!PyArg_ParseTuple(args, "K", &timestamp)) {
+		return NULL;
+	}
+
+	if (timestamp > UINT32_MAX) {
+		PyErr_SetString(PyExc_ValueError, "Time out of range");
+		return NULL;
+	}
+	status = dns_timestamp_to_nt_time(&nt, (uint32_t)timestamp);
+	if (!NT_STATUS_IS_OK(status)) {
+		PyErr_SetString(PyExc_ValueError, "Time out of range");
+		return NULL;
+	}
+	return Py_BuildValue("L", (long long) nt);
+}
+
+
 static PyMethodDef py_dsdb_dns_methods[] = {
 
 	{ "lookup", PY_DISCARD_FUNC_SIG(PyCFunction, py_dsdb_dns_lookup),
@@ -334,9 +410,18 @@ static PyMethodDef py_dsdb_dns_methods[] = {
 		METH_VARARGS, "Replace the DNS database entries for a DNS name"},
 	{ "replace_by_dn", (PyCFunction)py_dsdb_dns_replace_by_dn,
 		METH_VARARGS, "Replace the DNS database entries for a LDB DN"},
+	{ "records_match", (PyCFunction)py_dsdb_dns_records_match,
+	  METH_VARARGS|METH_KEYWORDS,
+	  "Decide whether two records match, according to dns update rules"},
 	{ "extract", (PyCFunction)py_dsdb_dns_extract,
 		METH_VARARGS, "Return the DNS database entry as a python structure from an Ldb.MessageElement of type dnsRecord"},
-	{ NULL }
+	{ "unix_to_dns_timestamp", (PyCFunction)py_dsdb_dns_unix_to_dns_timestamp,
+	  METH_VARARGS,
+	  "Convert a time.time() value to a dns timestamp (hours since 1601)"},
+	{ "dns_timestamp_to_nt_time", (PyCFunction)py_dsdb_dns_timestamp_to_nt_time,
+	  METH_VARARGS,
+	  "Convert a dns timestamp to an NTTIME value"},
+	{0}
 };
 
 static struct PyModuleDef moduledef = {
