@@ -45,6 +45,8 @@
 #include "../lib/tsocket/tsocket.h"
 #include "librpc/gen_ndr/ndr_winbind_c.h"
 #include "lib/messaging/irpc.h"
+#include "hdb.h"
+#include <kdc-audit.h>
 
 static krb5_error_code hdb_samba4_open(krb5_context context, HDB *db, int flags, mode_t mode)
 {
@@ -78,33 +80,187 @@ static krb5_error_code hdb_samba4_rename(krb5_context context, HDB *db, const ch
 	return HDB_ERR_DB_INUSE;
 }
 
-static krb5_error_code hdb_samba4_store(krb5_context context, HDB *db, unsigned flags, hdb_entry_ex *entry)
+static krb5_error_code hdb_samba4_store(krb5_context context, HDB *db, unsigned flags, hdb_entry *entry)
 {
 	return HDB_ERR_DB_INUSE;
 }
 
-static krb5_error_code hdb_samba4_remove(krb5_context context, HDB *db, krb5_const_principal principal)
+/*
+ * If we ever want kadmin to work fast, we might try and reopen the
+ * ldb with LDB_NOSYNC
+ */
+static krb5_error_code hdb_samba4_set_sync(krb5_context context, struct HDB *db, int set_sync)
 {
-	return HDB_ERR_DB_INUSE;
+	return 0;
+}
+
+static void hdb_samba4_free_entry_context(krb5_context context, struct HDB *db, hdb_entry *entry)
+{
+	/*
+	 * This function is now called for every HDB entry, not just those with
+	 * 'context' set, so we have to check that the context is not NULL.
+	*/
+	if (entry->context != NULL) {
+		/* this function is called only from hdb_free_entry().
+		 * Make sure we neutralize the destructor or we will
+		 * get a double free later when hdb_free_entry() will
+		 * try to call free_hdb_entry() */
+		talloc_set_destructor(entry->context, NULL);
+
+		/* now proceed to free the talloc part */
+		talloc_free(entry->context);
+	}
+}
+
+static int hdb_samba4_fill_fast_cookie(krb5_context context,
+				       struct samba_kdc_db_context *kdc_db_ctx)
+{
+	struct ldb_message *msg = ldb_msg_new(kdc_db_ctx);
+	int ldb_ret;
+
+	uint8_t secretbuffer[32];
+	struct ldb_val val = data_blob_const(secretbuffer,
+					     sizeof(secretbuffer));
+
+	if (msg == NULL) {
+		DBG_ERR("Failed to allocate msg for new fast cookie\n");
+		return LDB_ERR_OPERATIONS_ERROR;
+	}
+
+	/* Fill in all the keys with the same secret */
+	generate_secret_buffer(secretbuffer,
+			       sizeof(secretbuffer));
+
+	msg->dn = kdc_db_ctx->fx_cookie_dn;
+
+	ldb_ret = ldb_msg_add_value(msg, "secret", &val, NULL);
+
+	if (ldb_ret != LDB_SUCCESS) {
+		return ldb_ret;
+	}
+
+	ldb_ret = ldb_add(kdc_db_ctx->secrets_db,
+			  msg);
+	if (ldb_ret != LDB_SUCCESS) {
+		DBG_ERR("Failed to add fast cookie to ldb: %s\n",
+			ldb_errstring(kdc_db_ctx->secrets_db));
+	}
+	return ldb_ret;
+}
+
+static krb5_error_code hdb_samba4_fetch_fast_cookie(krb5_context context,
+						    struct samba_kdc_db_context *kdc_db_ctx,
+						    hdb_entry *entry_ex)
+{
+	krb5_error_code ret = SDB_ERR_NOENTRY;
+	TALLOC_CTX *mem_ctx;
+	struct ldb_result *res;
+	int ldb_ret;
+	struct sdb_entry_ex sdb_entry_ex = {};
+	const char *attrs[] = {
+		"secret",
+		NULL
+	};
+	const struct ldb_val *val;
+
+	mem_ctx = talloc_named(kdc_db_ctx, 0, "samba_kdc_fetch context");
+	if (!mem_ctx) {
+		ret = ENOMEM;
+		krb5_set_error_message(context, ret, "samba_kdc_fetch: talloc_named() failed!");
+		return ret;
+	}
+
+	/* search for CN=FX-COOKIE */
+	ldb_ret = ldb_search(kdc_db_ctx->secrets_db,
+			     mem_ctx,
+			     &res,
+			     kdc_db_ctx->fx_cookie_dn,
+			     LDB_SCOPE_BASE,
+			     attrs, NULL);
+
+	if (ldb_ret == LDB_ERR_NO_SUCH_OBJECT || res->count == 0) {
+
+		ldb_ret = hdb_samba4_fill_fast_cookie(context,
+						      kdc_db_ctx);
+
+		if (ldb_ret != LDB_SUCCESS) {
+			TALLOC_FREE(mem_ctx);
+			return HDB_ERR_NO_WRITE_SUPPORT;
+		}
+
+		/* search for CN=FX-COOKIE */
+		ldb_ret = ldb_search(kdc_db_ctx->secrets_db,
+				     mem_ctx,
+				     &res,
+				     kdc_db_ctx->fx_cookie_dn,
+				     LDB_SCOPE_BASE,
+				     attrs, NULL);
+
+		if (ldb_ret != LDB_SUCCESS || res->count != 1) {
+			TALLOC_FREE(mem_ctx);
+			return HDB_ERR_NOENTRY;
+		}
+	}
+
+	val = ldb_msg_find_ldb_val(res->msgs[0],
+				   "secret");
+	if (val == NULL || val->length != 32) {
+		TALLOC_FREE(mem_ctx);
+		return HDB_ERR_NOENTRY;
+	}
+
+
+	ret = krb5_make_principal(context,
+				  &sdb_entry_ex.entry.principal,
+				  KRB5_WELLKNOWN_ORG_H5L_REALM,
+				  KRB5_WELLKNOWN_NAME, "org.h5l.fast-cookie",
+				  NULL);
+	if (ret) {
+		TALLOC_FREE(mem_ctx);
+		return ret;
+	}
+
+	ret = samba_kdc_set_fixed_keys(context, kdc_db_ctx,
+				       val, &sdb_entry_ex);
+	if (ret != 0) {
+		return ret;
+	}
+
+	ret = sdb_entry_ex_to_hdb_entry_ex(context,
+					   &sdb_entry_ex,
+					   entry_ex);
+	sdb_free_entry(&sdb_entry_ex);
+	TALLOC_FREE(mem_ctx);
+
+	return ret;
 }
 
 static krb5_error_code hdb_samba4_fetch_kvno(krb5_context context, HDB *db,
 					     krb5_const_principal principal,
 					     unsigned flags,
 					     krb5_kvno kvno,
-					     hdb_entry_ex *entry_ex)
+					     hdb_entry *entry_ex)
 {
 	struct samba_kdc_db_context *kdc_db_ctx;
 	struct sdb_entry_ex sdb_entry_ex = {};
 	krb5_error_code code, ret;
+	uint32_t sflags;
 
 	kdc_db_ctx = talloc_get_type_abort(db->hdb_db,
 					   struct samba_kdc_db_context);
 
+	if (flags & HDB_F_GET_FAST_COOKIE) {
+		return hdb_samba4_fetch_fast_cookie(context,
+						    kdc_db_ctx,
+						    entry_ex);
+	}
+
+	sflags = (flags & SDB_F_HDB_MASK);
+
 	ret = samba_kdc_fetch(context,
 			      kdc_db_ctx,
 			      principal,
-			      flags,
+			      sflags,
 			      kvno,
 			      &sdb_entry_ex);
 	switch (ret) {
@@ -137,7 +293,7 @@ static krb5_error_code hdb_samba4_fetch_kvno(krb5_context context, HDB *db,
 }
 
 static krb5_error_code hdb_samba4_firstkey(krb5_context context, HDB *db, unsigned flags,
-					hdb_entry_ex *entry)
+					hdb_entry *entry)
 {
 	struct samba_kdc_db_context *kdc_db_ctx;
 	struct sdb_entry_ex sdb_entry_ex = {};
@@ -166,7 +322,7 @@ static krb5_error_code hdb_samba4_firstkey(krb5_context context, HDB *db, unsign
 }
 
 static krb5_error_code hdb_samba4_nextkey(krb5_context context, HDB *db, unsigned flags,
-				   hdb_entry_ex *entry)
+				   hdb_entry *entry)
 {
 	struct samba_kdc_db_context *kdc_db_ctx;
 	struct sdb_entry_ex sdb_entry_ex = {};
@@ -202,7 +358,7 @@ static krb5_error_code hdb_samba4_destroy(krb5_context context, HDB *db)
 
 static krb5_error_code
 hdb_samba4_check_constrained_delegation(krb5_context context, HDB *db,
-					hdb_entry_ex *entry,
+					hdb_entry *entry,
 					krb5_const_principal target_principal)
 {
 	struct samba_kdc_db_context *kdc_db_ctx;
@@ -211,7 +367,7 @@ hdb_samba4_check_constrained_delegation(krb5_context context, HDB *db,
 
 	kdc_db_ctx = talloc_get_type_abort(db->hdb_db,
 					   struct samba_kdc_db_context);
-	skdc_entry = talloc_get_type_abort(entry->ctx,
+	skdc_entry = talloc_get_type_abort(entry->context,
 					   struct samba_kdc_entry);
 
 	ret = samba_kdc_check_s4u2proxy(context, kdc_db_ctx,
@@ -238,7 +394,7 @@ hdb_samba4_check_constrained_delegation(krb5_context context, HDB *db,
 
 static krb5_error_code
 hdb_samba4_check_pkinit_ms_upn_match(krb5_context context, HDB *db,
-				     hdb_entry_ex *entry,
+				     hdb_entry *entry,
 				     krb5_const_principal certificate_principal)
 {
 	struct samba_kdc_db_context *kdc_db_ctx;
@@ -247,7 +403,7 @@ hdb_samba4_check_pkinit_ms_upn_match(krb5_context context, HDB *db,
 
 	kdc_db_ctx = talloc_get_type_abort(db->hdb_db,
 					   struct samba_kdc_db_context);
-	skdc_entry = talloc_get_type_abort(entry->ctx,
+	skdc_entry = talloc_get_type_abort(entry->context,
 					   struct samba_kdc_entry);
 
 	ret = samba_kdc_check_pkinit_ms_upn_match(context, kdc_db_ctx,
@@ -273,20 +429,20 @@ hdb_samba4_check_pkinit_ms_upn_match(krb5_context context, HDB *db,
 }
 
 static krb5_error_code
-hdb_samba4_check_s4u2self(krb5_context context, HDB *db,
-			  hdb_entry_ex *client_entry,
-			  hdb_entry_ex *server_target_entry)
+hdb_samba4_check_client_matches_target_service(krb5_context context, HDB *db,
+			  hdb_entry *client_entry,
+			  hdb_entry *server_target_entry)
 {
 	struct samba_kdc_entry *skdc_client_entry
-		= talloc_get_type_abort(client_entry->ctx,
+		= talloc_get_type_abort(client_entry->context,
 					struct samba_kdc_entry);
 	struct samba_kdc_entry *skdc_server_target_entry
-		= talloc_get_type_abort(server_target_entry->ctx,
+		= talloc_get_type_abort(server_target_entry->context,
 					struct samba_kdc_entry);
 
-	return samba_kdc_check_s4u2self(context,
-					skdc_client_entry,
-					skdc_server_target_entry);
+	return samba_kdc_check_client_matches_target_service(context,
+							     skdc_client_entry,
+							     skdc_server_target_entry);
 }
 
 static void reset_bad_password_netlogon(TALLOC_CTX *mem_ctx,
@@ -311,40 +467,79 @@ static void reset_bad_password_netlogon(TALLOC_CTX *mem_ctx,
 					irpc_handle, &req);
 }
 
-static krb5_error_code hdb_samba4_auth_status(krb5_context context, HDB *db,
-					      hdb_entry_ex *entry,
-					      struct sockaddr *from_addr,
-					      struct timeval *start_time,
-					      const char *original_client_name,
-					      const char *auth_type,
-					      int hdb_auth_status)
+static krb5_error_code hdb_samba4_audit(krb5_context context,
+					HDB *db,
+					hdb_entry *entry,
+					hdb_request_t r)
 {
 	struct samba_kdc_db_context *kdc_db_ctx = talloc_get_type_abort(db->hdb_db,
 									struct samba_kdc_db_context);
-
 	struct ldb_dn *domain_dn = ldb_get_default_basedn(kdc_db_ctx->samdb);
 	uint64_t logon_id = generate_random_u64();
+	heim_object_t auth_details_obj = NULL;
+	const char *auth_details = NULL;
+	char *etype_str = NULL;
+	heim_object_t hdb_auth_status_obj = NULL;
+	int hdb_auth_status;
+	heim_object_t pa_type_obj = NULL;
+	const char *pa_type = NULL;
+	struct auth_usersupplied_info ui;
+	size_t sa_socklen = 0;
+	int final_ret = 0;
+
+	hdb_auth_status_obj = heim_audit_getkv((heim_svc_req_desc)r, KDC_REQUEST_KV_AUTH_EVENT);
+	if (hdb_auth_status_obj == NULL) {
+		/* No status code found, so just return. */
+		return 0;
+	}
+
+	hdb_auth_status = heim_number_get_int(hdb_auth_status_obj);
+
+	pa_type_obj = heim_audit_getkv((heim_svc_req_desc)r, "pa");
+	if (pa_type_obj != NULL) {
+		pa_type = heim_string_get_utf8(pa_type_obj);
+	}
+
+	auth_details_obj = heim_audit_getkv((heim_svc_req_desc)r, KDC_REQUEST_KV_PKINIT_CLIENT_CERT);
+	if (auth_details_obj != NULL) {
+		auth_details = heim_string_get_utf8(auth_details_obj);
+	} else {
+		auth_details_obj = heim_audit_getkv((heim_svc_req_desc)r, KDC_REQUEST_KV_GSS_INITIATOR);
+		if (auth_details_obj != NULL) {
+			auth_details = heim_string_get_utf8(auth_details_obj);
+		} else {
+			heim_object_t etype_obj = heim_audit_getkv((heim_svc_req_desc)r, KDC_REQUEST_KV_PA_ETYPE);
+			if (etype_obj != NULL) {
+				int etype = heim_number_get_int(etype_obj);
+
+				krb5_error_code ret = krb5_enctype_to_string(r->context, etype, &etype_str);
+				if (ret == 0) {
+					auth_details = etype_str;
+				} else {
+					auth_details = "unknown enctype";
+				}
+			}
+		}
+	}
 
 	/*
 	 * Forcing this via the NTLM auth structure is not ideal, but
 	 * it is the most practical option right now, and ensures the
 	 * logs are consistent, even if some elements are always NULL.
 	 */
-	struct auth_usersupplied_info ui = {
+	ui = (struct auth_usersupplied_info) {
 		.was_mapped = true,
 		.client = {
-			.account_name = original_client_name,
+			.account_name = r->cname,
 			.domain_name = NULL,
 		},
 		.service_description = "Kerberos KDC",
-		.auth_description = "ENC-TS Pre-authentication",
-		.password_type = auth_type,
+		.auth_description = "Unknown Auth Description",
+		.password_type = auth_details,
 		.logon_id = logon_id
 	};
-	size_t sa_socklen = 0;
-	int final_ret = 0;
 
-	switch (from_addr->sa_family) {
+	switch (r->addr->sa_family) {
 	case AF_INET:
 		sa_socklen = sizeof(struct sockaddr_in);
 		break;
@@ -356,10 +551,10 @@ static krb5_error_code hdb_samba4_auth_status(krb5_context context, HDB *db,
 	}
 
 	switch (hdb_auth_status) {
-	case HDB_AUTHZ_SUCCESS:
+	case KDC_AUTH_EVENT_CLIENT_AUTHORIZED:
 	{
 		TALLOC_CTX *frame = talloc_stackframe();
-		struct samba_kdc_entry *p = talloc_get_type(entry->ctx,
+		struct samba_kdc_entry *p = talloc_get_type(entry->context,
 							    struct samba_kdc_entry);
 		struct netr_SendToSamBase *send_to_sam = NULL;
 
@@ -374,15 +569,12 @@ static krb5_error_code hdb_samba4_auth_status(krb5_context context, HDB *db,
 			reset_bad_password_netlogon(frame, kdc_db_ctx, send_to_sam);
 		}
 		talloc_free(frame);
-		break;
 	}
-	case HDB_AUTH_INVALID_SIGNATURE:
-		break;
-	case HDB_AUTH_CORRECT_PASSWORD:
-	case HDB_AUTH_WRONG_PASSWORD:
+	FALL_THROUGH;
+	default:
 	{
 		TALLOC_CTX *frame = talloc_stackframe();
-		struct samba_kdc_entry *p = talloc_get_type(entry->ctx,
+		struct samba_kdc_entry *p = talloc_get_type(entry->context,
 							    struct samba_kdc_entry);
 		struct dom_sid *sid
 			= samdb_result_dom_sid(frame, p->msg, "objectSid");
@@ -390,11 +582,12 @@ static krb5_error_code hdb_samba4_auth_status(krb5_context context, HDB *db,
 			= ldb_msg_find_attr_as_string(p->msg, "sAMAccountName", NULL);
 		const char *domain_name = lpcfg_sam_name(p->kdc_db_ctx->lp_ctx);
 		struct tsocket_address *remote_host;
+		const char *auth_description = NULL;
 		NTSTATUS status;
 		int ret;
 		bool rwdc_fallback = false;
 
-		ret = tsocket_address_bsd_from_sockaddr(frame, from_addr,
+		ret = tsocket_address_bsd_from_sockaddr(frame, r->addr,
 							sa_socklen,
 							&remote_host);
 		if (ret != 0) {
@@ -406,12 +599,86 @@ static krb5_error_code hdb_samba4_auth_status(krb5_context context, HDB *db,
 		ui.mapped.account_name = account_name;
 		ui.mapped.domain_name = domain_name;
 
-		if (hdb_auth_status == HDB_AUTH_WRONG_PASSWORD) {
+		if (pa_type != NULL) {
+			auth_description = talloc_asprintf(frame,
+							   "%s Pre-authentication",
+							   pa_type);
+			if (auth_description == NULL) {
+				auth_description = pa_type;
+			}
+		} else {
+			auth_description = "Unknown Pre-authentication";
+		}
+		ui.auth_description = auth_description;
+
+		if (hdb_auth_status == KDC_AUTH_EVENT_CLIENT_AUTHORIZED) {
+			/* This is the final sucess */
+			status = NT_STATUS_OK;
+		} else if (hdb_auth_status == KDC_AUTH_EVENT_VALIDATED_LONG_TERM_KEY) {
+			/*
+			 * This was only a pre-authentication success,
+			 * but we didn't reach the final
+			 * KDC_AUTH_EVENT_CLIENT_AUTHORIZED,
+			 * so consult the error code.
+			 */
+			if (r->error_code == 0) {
+				DBG_ERR("ERROR: VALIDATED_LONG_TERM_KEY "
+					"with error=0 => INTERNAL_ERROR\n");
+				status = NT_STATUS_INTERNAL_ERROR;
+				final_ret = KRB5KRB_ERR_GENERIC;
+				r->error_code = final_ret;
+			} else if (!NT_STATUS_IS_OK(p->reject_status)) {
+				status = p->reject_status;
+			} else {
+				status = krb5_to_nt_status(r->error_code);
+			}
+		} else if (hdb_auth_status == KDC_AUTH_EVENT_PREAUTH_SUCCEEDED) {
+			/*
+			 * This was only a pre-authentication success,
+			 * but we didn't reach the final
+			 * KDC_AUTH_EVENT_CLIENT_AUTHORIZED,
+			 * so consult the error code.
+			 */
+			if (r->error_code == 0) {
+				DBG_ERR("ERROR: PREAUTH_SUCCEEDED "
+					"with error=0 => INTERNAL_ERROR\n");
+				status = NT_STATUS_INTERNAL_ERROR;
+				final_ret = KRB5KRB_ERR_GENERIC;
+				r->error_code = final_ret;
+			} else if (!NT_STATUS_IS_OK(p->reject_status)) {
+				status = p->reject_status;
+			} else {
+				status = krb5_to_nt_status(r->error_code);
+			}
+		} else if (hdb_auth_status == KDC_AUTH_EVENT_CLIENT_TIME_SKEW) {
+			status = NT_STATUS_TIME_DIFFERENCE_AT_DC;
+		} else if (hdb_auth_status == KDC_AUTH_EVENT_WRONG_LONG_TERM_KEY) {
 			authsam_update_bad_pwd_count(kdc_db_ctx->samdb, p->msg, domain_dn);
 			status = NT_STATUS_WRONG_PASSWORD;
 			rwdc_fallback = kdc_db_ctx->rodc;
+		} else if (hdb_auth_status == KDC_AUTH_EVENT_CLIENT_LOCKED_OUT) {
+			status = NT_STATUS_ACCOUNT_LOCKED_OUT;
+			rwdc_fallback = kdc_db_ctx->rodc;
+		} else if (hdb_auth_status == KDC_AUTH_EVENT_CLIENT_NAME_UNAUTHORIZED) {
+			if (pa_type != NULL && strncmp(pa_type, "PK-INIT", strlen("PK-INIT")) == 0) {
+				status = NT_STATUS_PKINIT_NAME_MISMATCH;
+			} else {
+				status = NT_STATUS_ACCOUNT_RESTRICTION;
+			}
+			rwdc_fallback = kdc_db_ctx->rodc;
+		} else if (hdb_auth_status == KDC_AUTH_EVENT_PREAUTH_FAILED) {
+			if (pa_type != NULL && strncmp(pa_type, "PK-INIT", strlen("PK-INIT")) == 0) {
+				status = NT_STATUS_PKINIT_FAILURE;
+			} else {
+				status = NT_STATUS_GENERIC_COMMAND_FAILED;
+			}
+			rwdc_fallback = kdc_db_ctx->rodc;
 		} else {
-			status = NT_STATUS_OK;
+			DBG_ERR("Unhandled hdb_auth_status=%d => INTERNAL_ERROR\n",
+				hdb_auth_status);
+			status = NT_STATUS_INTERNAL_ERROR;
+			final_ret = KRB5KRB_ERR_GENERIC;
+			r->error_code = final_ret;
 		}
 
 		if (rwdc_fallback) {
@@ -419,27 +686,40 @@ static krb5_error_code hdb_samba4_auth_status(krb5_context context, HDB *db,
 			 * Forward the request to an RWDC in order
 			 * to give an authoritative answer to the client.
 			 */
-			ui.password_type = "Forwarding to RWDC";
+			auth_description = talloc_asprintf(frame,
+							   "%s,Forward-To-RWDC",
+							   ui.auth_description);
+			if (auth_description != NULL) {
+				ui.auth_description = auth_description;
+			}
 			final_ret = HDB_ERR_NOT_FOUND_HERE;
 		}
 
 		log_authentication_event(kdc_db_ctx->msg_ctx,
 					 kdc_db_ctx->lp_ctx,
-					 start_time,
+					 &r->tv_start,
 					 &ui,
 					 status,
 					 domain_name,
 					 account_name,
 					 sid);
+		if (final_ret == KRB5KRB_ERR_GENERIC && socket_wrapper_enabled()) {
+			/*
+			 * If we're running under make test
+			 * just panic
+			 */
+			DBG_ERR("Unexpected situation => PANIC\n");
+			smb_panic("hdb_samba4_audit: Unexpected situation");
+		}
 		TALLOC_FREE(frame);
 		break;
 	}
-	case HDB_AUTH_CLIENT_UNKNOWN:
+	case KDC_AUTH_EVENT_CLIENT_UNKNOWN:
 	{
 		struct tsocket_address *remote_host;
 		int ret;
 		TALLOC_CTX *frame = talloc_stackframe();
-		ret = tsocket_address_bsd_from_sockaddr(frame, from_addr,
+		ret = tsocket_address_bsd_from_sockaddr(frame, r->addr,
 							sa_socklen,
 							&remote_host);
 		if (ret != 0) {
@@ -448,11 +728,17 @@ static krb5_error_code hdb_samba4_auth_status(krb5_context context, HDB *db,
 			ui.remote_host = remote_host;
 		}
 
+		if (pa_type == NULL) {
+			pa_type = "AS-REQ";
+		}
+
+		ui.auth_description = pa_type;
+
 		/* Note this is not forwarded to an RWDC */
 
 		log_authentication_event(kdc_db_ctx->msg_ctx,
 					 kdc_db_ctx->lp_ctx,
-					 start_time,
+					 &r->tv_start,
 					 &ui,
 					 NT_STATUS_NO_SUCH_USER,
 					 NULL, NULL,
@@ -461,6 +747,9 @@ static krb5_error_code hdb_samba4_auth_status(krb5_context context, HDB *db,
 		break;
 	}
 	}
+
+	free(etype_str);
+
 	return final_ret;
 }
 
@@ -480,7 +769,7 @@ NTSTATUS hdb_samba4_create_kdc(struct samba_kdc_base_context *base_ctx,
 		return NT_STATUS_ERROR_DS_INCOMPATIBLE_VERSION;
 	}
 
-	*db = talloc(base_ctx, HDB);
+	*db = talloc_zero(base_ctx, HDB);
 	if (!*db) {
 		krb5_set_error_message(context, ENOMEM, "malloc: out of memory");
 		return NT_STATUS_NO_MEMORY;
@@ -500,13 +789,14 @@ NTSTATUS hdb_samba4_create_kdc(struct samba_kdc_base_context *base_ctx,
 	(*db)->hdb_dbc = NULL;
 	(*db)->hdb_open = hdb_samba4_open;
 	(*db)->hdb_close = hdb_samba4_close;
+	(*db)->hdb_free_entry_context = hdb_samba4_free_entry_context;
 	(*db)->hdb_fetch_kvno = hdb_samba4_fetch_kvno;
 	(*db)->hdb_store = hdb_samba4_store;
-	(*db)->hdb_remove = hdb_samba4_remove;
 	(*db)->hdb_firstkey = hdb_samba4_firstkey;
 	(*db)->hdb_nextkey = hdb_samba4_nextkey;
 	(*db)->hdb_lock = hdb_samba4_lock;
 	(*db)->hdb_unlock = hdb_samba4_unlock;
+	(*db)->hdb_set_sync = hdb_samba4_set_sync;
 	(*db)->hdb_rename = hdb_samba4_rename;
 	/* we don't implement these, as we are not a lockable database */
 	(*db)->hdb__get = NULL;
@@ -515,10 +805,10 @@ NTSTATUS hdb_samba4_create_kdc(struct samba_kdc_base_context *base_ctx,
 	(*db)->hdb__del = NULL;
 	(*db)->hdb_destroy = hdb_samba4_destroy;
 
-	(*db)->hdb_auth_status = hdb_samba4_auth_status;
+	(*db)->hdb_audit = hdb_samba4_audit;
 	(*db)->hdb_check_constrained_delegation = hdb_samba4_check_constrained_delegation;
 	(*db)->hdb_check_pkinit_ms_upn_match = hdb_samba4_check_pkinit_ms_upn_match;
-	(*db)->hdb_check_s4u2self = hdb_samba4_check_s4u2self;
+	(*db)->hdb_check_client_matches_target_service = hdb_samba4_check_client_matches_target_service;
 
 	return NT_STATUS_OK;
 }
